; ************************************************************************************************
; * 8-KANAL-DMX-DIMMER - V 0.3 - 01.01.2010													 	 *
; *																								 *
; * fr ATmega16 - Schaltung von Andr Zibell													 *
; ************************************************************************************************

.include "m16def.inc"

; ================================================================================================
; Register etc. definieren
; ------------------------------------------------------------------------------------------------

.def Null			= R2					; immer Null, fr 16-Bit-Operationen etc.
.def SREGbuf		= R3					; Puffer fr Status-Register, NUR ISR!
.def PWM_Count 		= R4					; Zhler fr SW-PWM
.def DMX_Base_L		= R5					; Low-Byte DMX-Startadresse
.def DMX_Base_H		= R6					; High-Byte DMX-Startadresse
.def ZL_Buf			= R7					; Puffer fr Z-Pointer DMX-Routine
.def ZH_Buf			= R8

.def temp0 			= R16					; Mehrzweck-Temp-Register
.def temp1 			= R17					; Mehrzweck-Temp-Register
.def temp2 			= R18					; Mehrzweck-Temp-Register
.def DMXstate		= R19					; Flag fr DMX-Routine
.def Blink			= R20					; Zhler fr Blink-Takt Status-LED

; ------------------------------------------------------------------------------------------------

.equ XTAL			= 16000000				; Prozessortakt 16 MHz (Bitte 16 MHz Quarz benutzen!)

.equ DMX_Daten		= 0x60					; Startadresse DMX Array im RAM
.equ DMX_Kanaele	= 8						; Zahl der DMX-Kanle	

.equ PWM_Port		= PortA					; Ausgabe-Port fr PWM

.equ Status_Port	= PortD					; Port fr Status-LED etc.
.equ Status_Pin		= PinD					; Pin fr Status etc.
.equ Status_LED		= 3						; Port-Pin fr Status-LED
.equ ERR_LED		= 4						; Port-Pin fr Error-LED
.equ RSDIR			= 2						; DMX senden oder empfangen
.equ ERR			= 5						; Alarm-Leitung

.equ DIP_Pin_L		= PinC					; Pin fr DIP 1-8
.equ DIP_Pin_H		= PinB					; Pin fr restliche DIPs
.equ DIP_9			= 0						; Port-Pin fr DIP 9
.equ DMX_Hold		= 1						; DIP fr Daten halten/lschen bei DMX-Ausfall

; ================================================================================================
; Interrupt-Vektor-Tabelle Mega16
; ------------------------------------------------------------------------------------------------

.CSEG

.org 0x000		rjmp    Start               ; Reset Handler
.org 0x002		reti						; External Interrupt Request 0
.org 0x004		reti						; External Interrupt Request 1
.org 0x006		rjmp    ISR_PWM				; Timer/Counter2 Compare Match
.org 0x008		reti						; Timer/Counter2 Overflow
.org 0x00A		reti						; Timer/Counter1 Capture Event
.org 0x00C		reti						; Timer/Counter1 Compare Match A
.org 0x00E		reti						; Timer/Counter1 Compare Match B
.org 0x010		reti						; Timer/Counter1 Overflow
.org 0x012		reti					    ; Timer/Counter0 Overflow
.org 0x014		reti						; Serial Transfer Complete
.org 0x016		rjmp	ISR_Byte_Received	; Rx Complete
.org 0x018		reti						; Data Register Empty
.org 0x01A		reti						; Tx Complete
.org 0x01C		reti						; ADC Complete
.org 0x01E		reti						; EEPROM Ready
.org 0x020		reti						; Analog Comparator
.org 0x022		reti						; Two-Wire Serial Interface
.org 0x024		reti						; External Interrupt Request 2
.org 0x026		reti						; Timer/Counter0 Compare Match
.org 0x028		reti						; Store Program Memory Ready

; ================================================================================================
; Startup, Stackpointer und Ports setzen, etc. Manuell! - hier keine Variablen!
; ------------------------------------------------------------------------------------------------

.org 0x030

Start:

		clr		Null						; "Null" auf Null setzen ;o)

		ldi 	temp0, LOW(RAMEND)			; LOW-Byte der obersten RAM-Adresse
		out 	SPL, temp0
		ldi 	temp0, HIGH(RAMEND) 		; HIGH-Byte der obersten RAM-Adresse
		out 	SPH, temp0

		ser		temp0						; Port A Ausgang (PWM)
		out		DDRA, temp0
		clr		temp0						; Port B und C Eingang (DIPs)
		out		DDRB, temp0
		out		DDRC, temp0

		ldi		temp0, 0b11011111			; div. Ein- und Ausgnge (LEDs, Error, RSDIR, etc.)
		out		DDRD, temp0

		sbi		Status_Port, ERR			; Pullup fr Alarm-Leitung

		ser		temp0						; Pullups fr DIP-Schalter
		out		PortB, temp0
		out		PortC, temp0

		rcall	SUB_DMX_Init				; DMX initialisieren
		rcall	SUB_PWM_Init				; PWM initialisieren

		sei									; Interrupts aktivieren

; ================================================================================================
; Hauptschleife, Reaktion bei fehlendem DMX-Signal
; ------------------------------------------------------------------------------------------------

Main:

		cpi		Blink, 128					; Timeout erreicht...?
		brlo	Main						; Nein, nix weiter machen

		sbic	DIP_Pin_H, DMX_Hold			; Schalter fr DMX Hold an...?
		rjmp	Main						; Ja, nix weiter machen

		rcall	SUB_Clear_DMXRAM			; Nein, Werte im RAM auf 0 setzen

		rjmp	Main

; ================================================================================================
; ISR fr 8-Kanal-Software-PWM, Fehlerberwachung und Status-LED
;
; benutzte Register      : temp0, temp1, YL, YH, werden gesichert
; gem. benutzte Register : Blink
; reservierte Register   : PWM_Count
; ------------------------------------------------------------------------------------------------

ISR_PWM:
 
        in   	SREGbuf, SREG				; Register sichern
        push 	temp1
		push	temp0		
		push	YL
		push	YH

		sbic	Status_Pin, ERR				; erst mal schauen, ob Alarm vorliegt
		rjmp	PWM_weiter					; nein, ganz normal weiter mit PWM

PWM_Error:

		out		PWM_Port, Null				; Fehler gemeldet, also Ausgnge aus
		sbi		Status_Port, ERR_LED		; Error-LED an
		rjmp	PWM_Error					; und "aufhngen"


PWM_weiter:

        ldi     YH, high(DMX_Daten)			; Startadresse DMX-Daten im RAM
        ldi     YL, low(DMX_Daten)

        ser     temp0						; Grundzustand alle LEDs an (1 = LED an, 0 = LED aus)

        inc     PWM_Count					; den PWM Zhler weiterzhlen

        ld      temp1, Y+               	; Wert fr Kanal 1 aus dem RAM holen
        cp      PWM_Count, temp1        	; Ist der Grenzwert fr Kanal 1 erreicht?
        brlo    PWM_2                   	; nein, weiter zu Kanal 2
        cbr	    temp0, (1<<0)       		; ja, LED Kanal 1 aus

PWM_2:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_3
        cbr	    temp0, (1<<1)

PWM_3:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_4
        cbr	    temp0, (1<<2)

PWM_4:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_5
        cbr	    temp0, (1<<3)

PWM_5:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_6
        cbr	    temp0, (1<<4)

PWM_6:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_7
        cbr	    temp0, (1<<5)

PWM_7:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_8
        cbr	    temp0, (1<<6)

PWM_8:

        ld      temp1, Y+               	; und so weiter.....
        cp      PWM_Count, temp1
        brlo    PWM_Out
        cbr	    temp0, (1<<7)

PWM_Out:

        out     PWM_Port, temp0				; Neue Bitbelegung fr PWM ausgeben

; ------------------------------------------------------------------------------------------------
; Status-LED blinken lassen
; ------------------------------------------------------------------------------------------------

		tst		PWM_Count					; 1 Zyklus durch?
		brne	PWM_Ende					; nein, Ende

		inc		Blink						; ja, Blink-Zhler 1 weiter (244x pro Sekunde)
		cpi		Blink, 255					; Bei Overflow auf 
		brne	Blink_weiter
		ldi		Blink, 128					; 128 setzen, also weiter blinken, aber Timeout bleibt

Blink_weiter:

		sbi		Status_Port, Status_LED		; Status-LED erst mal an
		sbrc	Blink, 6					; 7. Bit Testen -> ca. 2 Hz Blink-Takt
		cbi		Status_Port, Status_LED		; 7. Bit gesetzt -> LED aus (Blinken)

; ------------------------------------------------------------------------------------------------

PWM_Ende:

        pop		YH						; Register wiederherstellen
		pop		YL
		pop		temp0
		pop 	temp1
        out 	SREG, SREGbuf	
		
		reti

; ================================================================================================
; ISR fr DMX-Empfang
;
; benutzte Register      : temp0, temp1, ZL, ZH, werden gesichert
; gem. benutzte Register : Blink
; reservierte Register   : DMXstate, ZL_Buf, ZH_Buf
; ------------------------------------------------------------------------------------------------

ISR_Byte_Received:

		in      SREGbuf, SREG				; Register sichern, Z aus Puffer holen
		push	temp1
		push	temp0		
		push	ZL
		push	ZH
		mov		ZL, ZL_Buf
		mov		ZH, ZH_Buf

		cbi		Status_Port, RSDIR			; RS-485-Treiber auf Empfang

		in		temp0,UCSRA					; USART Status und Daten holen
		in		temp1,UDR

		sbrc	temp0,FE					; Frame-Error: neuer Durchgang
		rjmp	dmx_frame_error	
		
		cpi		DMXstate, 1   				; DMX-Status 1: Startbyte testen
		breq	dmx_startbyte

  		cpi 	DMXstate, 2   				; DMX-Status 2: Startadresse abfragen
		breq    dmx_startadr
  		
 		cpi 	DMXstate, 3					; DMX-Status 3: Byte empfangen und in's RAM
  		brsh	dmx_handle_byte

dmx_finish:

		mov		ZL_Buf, ZL
		mov		ZH_Buf, ZH
		pop		ZH
		pop		ZL
		pop		temp0
		pop		temp1
		out     SREG, SREGbuf	
		reti			

dmx_startbyte:

		tst 	temp1	 					; Byte = 0 --> Startbyte, sonst Fehler
		brne 	dmx_err
		inc  	DMXstate

		rcall	SUB_Get_Adress				; DIP-Schalter abfragen
		mov 	ZL, DMX_Base_L 				; Startadresse aus den Registern holen
		mov 	ZH, DMX_Base_H

		rjmp	dmx_finish

dmx_startadr:

		sbiw	ZH:ZL, 1					; Adresse runterzhlen
		brne	dmx_finish					; Startadresse noch nicht erreicht

		ldi 	ZH, high(DMX_Daten)
		ldi		ZL, low(DMX_Daten)
		inc  	DMXstate
		
dmx_handle_byte:
			
		clr		Blink						; Blink-Zhler resetten

		st		Z+, temp1

		cpi		ZL, low(DMX_Daten+DMX_Kanaele)
		brne	dmx_finish
		cpi		ZH, high(DMX_Daten+DMX_Kanaele)
		brne	dmx_finish	
		clr		DMXstate

		rjmp	dmx_finish
	
dmx_frame_error:

		ldi 	DMXstate, 1					; Frame-Error --> neuer Durchlauf
		cbi		UCSRA,FE					; Flag lschen, falls lngere Pause
		rjmp	dmx_finish

dmx_err:

	 	clr 	DMXstate					; auf nchsten FE warten			
		rjmp	dmx_finish

; ================================================================================================
; DIP-Schalter abfragen (Startadresse)
; ------------------------------------------------------------------------------------------------

SUB_Get_Adress:

		push	temp0

		in	 	DMX_Base_L, DIP_Pin_L 		; Startadresse Low-Byte von DIP 1-9 holen
		com     DMX_Base_L					; invertieren, da DIPs gegen GND schalten
		clr		DMX_Base_H
		sbis	DIP_Pin_H, DIP_9			; 9. Bit von DIP 9 holen
		inc		DMX_Base_H

		in 		temp0, DIP_Pin_H			; DIPs fr Subadresse holen
		com     temp0						; invertieren, da DIPs gegen GND schalten
		andi	temp0, 0b11110000			; Sub-Adresse liegt in Bit 4-7
		lsr		temp0						; ergibt Adresse * 8

		add		DMX_Base_L, temp0			; Subadresse zu Startadresse addieren
		adc		DMX_Base_H, Null			; in 16 Bit

		pop		temp0
		ret

; ================================================================================================
; Timer fr PWM initialisieren
; ------------------------------------------------------------------------------------------------

SUB_PWM_Init:

		ldi		temp0, 128					; alle 128 Takte Timer-Interrupt -> 488 Hz PWM
		out		OCR2, temp0

		ldi     temp0, (1<<OCIE2)      		; OCIE2: Interrupt bei Timer2 Compare Match an
		out     TIMSK, temp0

		ldi     temp0, (1<<WGM21) | (1<<CS20)  	; CTC-Mode, Prescaler auf 1, Timer starten
		out		TCCR2, temp0

		ret

; ================================================================================================
; DMX Initialisieren
; ------------------------------------------------------------------------------------------------

SUB_DMX_Init:

    	ldi 	temp0, ((XTAL/4000000)-1) 	; Usart auf 250 kbaud setzen
    	out 	UBRRL, temp0
    	clr 	temp0
    	out 	UBRRH, temp0

    	ldi 	temp0, (1<<URSEL)|(3<<UCSZ0)|(1<<USBS) 	; 8 Datenbits, 2 Stoppbits
    	out 	UCSRC, temp0			

		sbi 	UCSRB, RXEN					; Receiver enable
		sbi 	UCSRB, RXCIE				; RX Complete Int. enable

		rcall	SUB_Clear_DMXRAM			; alle Kanle im RAM auf 0 setzen

		ret

; ================================================================================================
; DMX-Daten im RAM lschen
; ------------------------------------------------------------------------------------------------

SUB_Clear_DMXRAM:

		push	ZL
		push	ZH
		push	temp0

		ldi		ZH, high(DMX_Daten)			; alle Kanle im RAM auf 0 setzen
		ldi		ZL, low(DMX_Daten)
		ldi 	temp0, DMX_Kanaele

Clear_Byte:	

		st 		Z+,Null
		dec		temp0
		brne	Clear_Byte

		pop		temp0
		pop		ZH
		pop		ZL
		ret

; ================================================================================================
